import pytest
from brownie import chain
from tests.fixtures import blocks_for_contract, mutate_tokens
from y.classes.common import ERC20
from y.networks import Network

ATOKENS = {
    Network.Mainnet: [
        "0xfC1E690f61EFd961294b3e1Ce3313fBD8aa4f85d",
        "0x4DA9b813057D04BAef4e5800E36083717b4a0341",
        "0x9bA00D6856a4eDF4665BcA2C2309936572473B7E",
        "0x71fc860F7D3A592A4a98740e39dB31d25db65ae8",
        "0x625aE63000f46200499120B906716420bd059240",
        "0xE1BA0FB44CCb0D11b80F92f4f8Ed94CA3fF51D00",
        "0x3a3A65aAb0dd2A17E3F1947bA16138cd37d08c04",
        "0x9D91BE44C06d373a8a226E1f3b146956083803eB",
        "0x7deB5e830be29F91E298ba5FF1356BB7f8146998",
        "0x6Ee0f7BB50a54AB5253dA0667B0Dc2ee526C30a8",
        "0x12e51E77DAAA58aA0E9247db7510Ea4B46F9bEAd",
        "0x3Ed3B47Dd13EC9a98b44e6204A523E766B225811",
        "0x9ff58f4fFB29fA2266Ab25e75e2A8b3503311656",
        "0x030bA81f1c18d280636F32af80b9AAd02Cf0854e",
        "0x5165d24277cD063F5ac44Efd447B27025e888f37",
        "0xDf7FF54aAcAcbFf42dfe29DD6144A69b629f8C9e",
        "0xB9D7CB55f463405CDfBe4E90a6D2Df01C2B92BF1",
        "0xFFC97d72E13E01096502Cb8Eb52dEe56f74DAD7B",
        "0x05Ec93c0365baAeAbF7AefFb0972ea7ECdD39CF1",
        "0xA361718326c15715591c299427c62086F69923D9",
        "0x028171bCA77440897B824Ca71D1c56caC55b68A3",
        "0xaC6Df26a590F08dcC95D5a4705ae8abbc88509Ef",
        "0x39C6b3e42d6A679d7D776778Fe880BC9487C2EDA",
        "0xa06bC25B5805d5F8d82847D191Cb4Af5A3e873E0",
        "0xa685a61171bb30d4072B338c80Cb7b2c865c873E",
        "0xc713e5E149D5D0715DcD1c156a020976e7E56B88",
        "0xCC12AbE4ff81c9378D670De1b57F8e0Dd228D77a",
        "0x35f6B052C598d933D69A4EEC4D04c73A191fE6c2",
        "0x6C5024Cd4F8A59110119C56f8933403A539555EB",
        "0x101cc05f4A51C0319f570d5E146a8C625198e636",
        "0xBcca60bB61934080951369a648Fb03DF4F96263C",
        "0x8dAE6Cb04688C62d939ed9B68d32Bc62e49970b1",
        "0xD37EE7e4f452C6638c96536e68090De8cBcdb583",
        "0x272F97b7a56a387aE942350bBC7Df5700f8a4576",
        "0xF256CC7847E919FAc9B808cC216cAc87CCF2f47a",
        "0xc9BC48c72154ef3e5425641a3c747242112a46AF",
        "0x1E6bb68Acec8fefBD87D192bE09bb274170a0548",
        "0x2e8F4bdbE3d47d7d7DE490437AeA9915D930F1A3",
        "0x6F634c6135D2EBD550000ac92F494F9CB8183dAe",
        "0xd4937682df3C8aEF4FE912A96A74121C0829E664",
        "0x683923dB55Fead99A79Fa01A27EeC3cB19679cC3",
        "0xf9Fb4AD91812b704Ba883B11d2B576E890a6730A",
        "0x79bE75FFC64DD58e66787E4Eae470c8a1FD08ba4",
        "0xd24946147829DEaA935bE2aD85A3291dbf109c80",
        "0x17a79792Fe6fE5C95dFE95Fe3fCEE3CAf4fE4Cb7",
        "0x13B2f6928D7204328b0E8E4BCd0379aA06EA21FA",
        "0x9303EabC860a743aABcc3A1629014CaBcc3F8D36",
        "0xc58F53A8adff2fB4eb16ED56635772075E2EE123",
        "0xe59d2FF6995a926A574390824a657eEd36801E55",
        "0xA1B0edF4460CC4d8bFAA18Ed871bFF15E5b57Eb4",
        "0xE340B25fE32B1011616bb8EC495A4d503e322177",
        "0x0ea20e7fFB006d4Cfe84df2F72d8c7bD89247DB0",
        "0xb8db81B84d30E2387de0FF330420A4AAA6688134",
        "0x370adc71f67f581158Dc56f539dF5F399128Ddf9",
        "0xA9e201A4e269d6cd5E9F0FcbcB78520cf815878B",
        "0x38E491A71291CD43E8DE63b7253E482622184894",
        "0x3D26dcd840fCC8e4B2193AcE8A092e4a65832F9f",
        "0x391E86e2C002C70dEe155eAceB88F7A3c38f5976",
        "0x2365a4890eD8965E564B7E2D27C38Ba67Fec4C6F",
        "0x5394794Be8b6eD5572FCd6b27103F46b5F390E8f",
        "0x358bD0d980E031E23ebA9AA793926857703783BD",
        "0xd109b2A304587569c84308c55465cd9fF0317bFB",
        "0xd145c6ae8931ed5Bca9b5f5B7dA5991F5aB63B5c",
        "0xCa5DFDABBfFD58cfD49A9f78Ca52eC8e0591a3C5",
        "0x2394eAc76A0B7992FB8aa0F0C126317Af76Ae707",
        "0x0EDf29d931b68D673bdCDD35B20e26eEC0532F25",
        "0x762E5982EB0D944A9800D7caBAF640464E892C91",
        "0x0A6172fb19E85DBae18898eA2f7a596B5681f1eE",
    ]
}.get(chain.id, [])

ATOKENS = mutate_tokens(ATOKENS)

@pytest.mark.parametrize('token',ATOKENS)
def test_erc20(token):
    token = ERC20(token)
    assert token.contract, f'Cannot fetch contract for token {token}'
    assert token.build_name, f'Cannot fetch build name for token {token}'
    assert token.symbol, f'Cannot fetch symbol for token {token}'
    assert token.name, f'Cannot fetch name for token {token}'
    assert token.decimals, f'Cannot fetch decimals for token {token}'
    assert token.scale, f'Cannot fetch scale for token {token}'
    assert 10 ** token.decimals == token.scale, f'Incorrect scale fetched for token {token}'
    assert token.total_supply() is not None, f'Cannot fetch total supply for token {token}'
    assert token.total_supply_readable() is not None, f'Cannot fetch total supply readable for token {token}'
    assert token.total_supply() / token.scale == token.total_supply_readable(), f'Incorrect total supply readable for token {token}'
    assert token.price(), f'Cannot fetch price for token {token}'

    for block in blocks_for_contract(token.address):
        assert token._decimals(block), f'Cannot fetch decimals for token {token}'
        assert token._scale(block), f'Cannot fetch scale for token {token}'
        assert token.total_supply(block) is not None, f'Cannot fetch total supply for token {token}'
        assert token.total_supply_readable(block) is not None, f'Cannot fetch total supply readable for token {token}'
        assert token.total_supply(block) / token._scale(block) == token.total_supply_readable(block), f'Incorrect total supply readable for token {token}'
        assert token.price(block), f'Cannot fetch price for token {token}'



import logging
import threading
from functools import lru_cache
from typing import Any, Callable, List, Optional, Union

import brownie
from brownie import chain, web3
from brownie.exceptions import CompilerError, ContractNotFound
from brownie.typing import AccountsType
from multicall import Call, Multicall

from y import convert
from y.decorators import log
from y.exceptions import (ContractNotVerified, MessedUpBrownieContract,
                          NodeNotSynced, call_reverted, contract_not_verified)
from y.interfaces.ERC20 import ERC20ABI
from y.networks import Network
from y.typing import Address, AnyAddressType, Block
from y.utils.cache import memory

logger = logging.getLogger(__name__)


def Contract_erc20(address: AnyAddressType) -> brownie.Contract:
    address = convert.to_address(address)
    return Contract.from_abi('ERC20',address,ERC20ABI)


def Contract_with_erc20_fallback(address: AnyAddressType) -> brownie.Contract:
    if type(address) is brownie.Contract:
        return address
    address = convert.to_address(address)
    try:
        return Contract(address)
    except (ContractNotVerified,MessedUpBrownieContract):
        return Contract_erc20(address)


@log(logger)
@memory.cache()
def contract_creation_block(address: AnyAddressType) -> int:
    """
    Determine the block when a contract was created using binary search.
    NOTE Requires access to historical state. Doesn't account for CREATE2 or SELFDESTRUCT.
    """
    address = convert.to_address(address)
    logger.info("contract creation block %s", address)
    height = chain.height
    lo, hi = 0, height

    if hi == 0:
        raise NodeNotSynced(f'''
            `chain.height` returns 0 on your node, which implies it is not fully synced.
            You can only run this function on a fully synced node.''')
    
    while hi - lo > 1:
        mid = lo + (hi - lo) // 2
        try:
            if web3.eth.get_code(address, block_identifier=mid):
                hi = mid
            else:
                lo = mid
        except ValueError as e:
            if 'missing trie node' in str(e):
                logger.critical('missing trie node, `contract_creation_block` may output a higher block than actual. Please try again using an archive node.')
            elif 'Server error: account aurora does not exist while viewing' in str(e):
                logger.critical(str(e))
            else:
                raise
            lo = mid
    return hi if hi != height else None

# cached Contract instance, saves about 20ms of init time
_contract_lock = threading.Lock()

@lru_cache
class Contract(brownie.Contract):
    def __init__(
        self, 
        address: AnyAddressType, 
        *args: Any, 
        owner: Optional[AccountsType] = None, 
        require_success: bool = True, 
        **kwargs: Any
        ) -> None:
        
        address = convert.to_address(address)
        
        with _contract_lock:
            try:
                try:
                    super().__init__(address, *args, owner=owner, **kwargs)
                    self._verified = True
                except AttributeError as e:
                    if "'UsingForDirective' object has no attribute 'typeName'" in str(e):
                        raise MessedUpBrownieContract(address, str(e))
                    raise
                except CompilerError as e:
                    raise MessedUpBrownieContract(address, str(e))
                except ConnectionError as e:
                    if '{"message":"Something went wrong.","result":null,"status":"0"}' in str(e):
                        if web3.eth.get_code(address): raise ContractNotVerified(address)
                        else: raise ContractNotFound(address)
                    raise
                except IndexError as e:
                    if 'list index out of range' in str(e): raise MessedUpBrownieContract(address, str(e))
                    elif "pop from an empty deque" in str(e): raise MessedUpBrownieContract(address, str(e))
                    else: raise
                except ValueError as e:
                    if contract_not_verified(e): raise ContractNotVerified(f'{address} on {Network.printable()}')
                    elif "Unknown contract address:" in str(e): raise ContractNotVerified(str(e)) # avax snowtrace
                    elif "invalid literal for int() with base 16" in str(e): raise MessedUpBrownieContract(address, str(e))
                    else: raise
            except (ContractNotFound, ContractNotVerified, MessedUpBrownieContract) as e:
                if require_success: raise
                if type(e) == ContractNotVerified: self._verified = False
                else: self._verified = None

    def has_method(self, method: str, return_response: bool = False) -> Union[bool,Any]:
        return has_method(self.address, method, return_response=return_response)

    def has_methods(
        self, 
        methods: List[str],
        func: Union[any, all] = all
    ) -> bool:
        return has_methods(self.address, methods, func)

    def build_name(self, return_None_on_failure: bool = False) -> Optional[str]:
        return build_name(self.address, return_None_on_failure=return_None_on_failure)


@log(logger)
@memory.cache()
def is_contract(address: AnyAddressType) -> bool:
    '''
    Checks to see if the input address is a contract. Returns `True` if:
    - The address is not and has never been a contract
    - The address used to be a contract but has self-destructed
    '''
    address = convert.to_address(address)
    return web3.eth.get_code(address) != '0x'

@log(logger)
@memory.cache()
def has_method(address: Address, method: str, return_response: bool = False) -> Union[bool,Any]:
    '''
    Checks to see if a contract has a `method` view method with no inputs.
    `return_response=True` will return `response` in bytes if `response` else `False`
    '''
    address = convert.to_address(address)
    try:
        response = Call(address, [method], [['key', None]])()['key']
    except Exception as e:
        if call_reverted(e):
            return False
        raise
    
    if response is None:
        return False
    if return_response:
        return response
    return True

@log(logger)
@memory.cache()
def has_methods(
    address: AnyAddressType, 
    methods: List[str],
    func: Callable = all # Union[any, all]
) -> bool:
    '''
    Checks to see if a contract has each view method (with no inputs) in `methods`.
    Pass `at_least_one=True` to only verify a contract has at least one of the methods.
    '''

    assert func in [all, any], '`func` must be either `any` or `all`'

    address = convert.to_address(address)
    calls = [Call(address, [method], [[method, None]]) for method in methods]
    try:
        response = Multicall(calls, require_success=False)().values()
        return func([False if call is None else True for call in response])
    except Exception as e:
        if not call_reverted(e): raise # and not out_of_gas(e): raise
        # Out of gas error implies one or more method is state-changing.
        # If `func == all` we return False because `has_methods` is only supposed to work for public view methods with no inputs
        # If `func == any` maybe one of the methods will work without "out of gas" error
        return False if func == all else any(has_method(address, method) for method in methods)


@log(logger)
def probe(
    address: AnyAddressType, 
    methods: List[str],
    block: Optional[Block] = None,
    return_method: bool = False
) -> Any:
    address = convert.to_address(address)
    calls = [Call(address, [method], [[method, None]]) for method in methods]
    results = Multicall(calls, block_id=block, require_success=False)()
    results = [(method, result) for method, result in results.items() if result is not None]
    assert len(results) in [1,0], '`probe` returned multiple results. Must debug'
    if len(results) == 1:
        method, result = results[0]
        if not return_method:
            return result
        else:
            return method, result


@log(logger)
@memory.cache()
def build_name(address: AnyAddressType, return_None_on_failure: bool = False) -> str:
    try:
        contract = Contract(address)
        return contract.__dict__['_build']['contractName']
    except ContractNotVerified:
        if not return_None_on_failure:
            raise
        return None
